import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.Timer;

import mensajesSIP.ACKMessage;
import mensajesSIP.ByeMessage;
import mensajesSIP.InviteMessage;
import mensajesSIP.NotFoundMessage;
import mensajesSIP.OKMessage;
import mensajesSIP.RegisterMessage;
import mensajesSIP.SDPMessage;
import mensajesSIP.SIPException;
import mensajesSIP.SIPMessage;


public class UA {

	private  DatagramSocket socket;
	private  InetAddress ipProxy;
	private  int puertoUa, puertoProxy;
	private  String usuarioSip;
	private  boolean debug;
	private InterpreteOrden interprete;
	private HiloEscucha hilosocket;
	private Scanner lector;
	private int estado;
	private static final int REGISTRANDO=0;
	private static final int INVITANDO=1;
	private static final int SIN_ESTADO=-1;
	
	private String usuarioSIPotro;
	public static  final int MAXFORWARD=70;
	public static final String dominio="dominio";
	
	public DatagramSocket getSocket() {
		return socket;
	}
	public InetAddress getIpProxy() {
		return ipProxy;
	}
	public int getPuertoUa() {
		return puertoUa;
	}
	public int getPuertoProxy() {
		return puertoProxy;
	}
	public String getUsuarioSip() {
		return usuarioSip;
	}
	public int getEstado(){
		return this.estado;
	}
	public void setEstado(int estado){
		this.estado=estado;
	}
	public synchronized Scanner getLector(){
		return lector;
	}
	public UA(){
		this.lector=new Scanner(System.in);
	}
	public static void main(String[] args) throws SocketException{
		UA cliente=new UA();
		cliente.setEstado(SIN_ESTADO);
		if(!cliente.parsearParametros(args)){
			System.out.println("Error en los parametros");
			System.out.println("Parametros: usuarioSIP puertoescuchaUA IPProxy puertoescuchaProxy debug(true/false)");
			return;
		}
		
		cliente.socket=new DatagramSocket(cliente.puertoUa);//cliente escuchando en puerto Ua y dir iPproxy
		if(!cliente.registrarse()){
			System.out.println("Error en el registro");
			return;
		}else{
			cliente.hilosocket=new HiloEscucha(cliente);
			cliente.hilosocket.start();
			cliente.iniciarInterprete();
			
		}
		
		
		
	} 
	public void iniciarInterprete() {
		interprete=new InterpreteOrden(this);
		interprete.start();
	}
	public void terminarInterprete(){
		reiniciaScanner();
		interprete.interrupt();
	}
	public InterpreteOrden getInterprete(){
		return interprete;
	}
	
	public boolean enviarMensaje(SIPMessage msg){
		byte[] mensajeBytes=msg.toStringMessage().getBytes();
		DatagramPacket mensaje=new DatagramPacket(mensajeBytes,mensajeBytes.length,ipProxy,puertoProxy);
		
		try {
			socket.send(mensaje);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} 
		return true;
	}
	public boolean ejecutarOrdenInvite(String orden) {
		
		String[] vorden=orden.split(" ");
		if(vorden.length!=2){
			System.out.println("Error en el formato de invate");
			System.out.println("Formato:INVITE usuarioSIPDestino");
			return false;
		}else{
			usuarioSIPotro=vorden[1];
			double numaleatorio=Math.round(Math.random()*1000000);

			InviteMessage mensajeInvite;
			ArrayList<String> vias=getViasInicial();

			mensajeInvite=new InviteMessage();
			
			mensajeInvite.setVias(vias );
			mensajeInvite.setMaxForwards(MAXFORWARD);
			mensajeInvite.setDestination("sip:"+usuarioSIPotro+"@"+dominio);
			mensajeInvite.setFromUri("sip:"+usuarioSip);
			mensajeInvite.setFromName(usuarioSip.substring(0, usuarioSip.indexOf('@')));
			mensajeInvite.setToUri("sip:"+usuarioSIPotro+"@"+dominio);
			mensajeInvite.setToName(usuarioSIPotro);
			mensajeInvite.setCallId(Double.toString(numaleatorio));
			mensajeInvite.setcSeqNumber("2 INVITE\n");
			mensajeInvite.setContact("localhost:"+puertoUa);
			mensajeInvite.setContentLength(142);
			mensajeInvite.setContentType("application/sdp");
			
			SDPMessage sdp=new SDPMessage();
			ArrayList<Integer>opcionesSdp=new ArrayList<Integer>();
			opcionesSdp.add(96);
			opcionesSdp.add(97);
			opcionesSdp.add(98);
			sdp.setOptions(opcionesSdp);
			try {
				sdp.setIp(InetAddress.getLocalHost().getHostAddress().toString());
			} catch (UnknownHostException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				sdp.setIp("localhost");
			}
			sdp.setPort(puertoUa);
			mensajeInvite.setSdp(sdp);
			if(debug){
				System.out.println("MensajeInvite-----------------------------\n"
				+mensajeInvite.toStringMessage()
				+ "\n---------------------------------------------");
			}	
			return enviarMensaje(mensajeInvite);//envio mensaje y devuelvo true or false segun envia o no
		}
	}
	public ArrayList<String> getViasInicial(){
		ArrayList<String> vias=new ArrayList<String>();
		try {
			vias.add(InetAddress.getLocalHost().getHostAddress().toString()+":"+puertoUa+"\n");
		} catch (UnknownHostException e1) {
			// TODO Auto-generated catch block
			vias.add("localhost"+":"+puertoUa+"\n");

			e1.printStackTrace();
		}
		return vias;
	}
	
	public boolean registrarse(){
		this.estado=REGISTRANDO;
		//una vez creado el mensaje de registro y tras convertirlo en un array de bytes, 
		//metere ese mensaje en el datagramPacket para enviarlo
		try {
			byte[] datosRecibidos=new byte[1024];
			SIPMessage mensajeSip;
			/*if(!enviarRegistro()){
				
				return false;
			}*/
			
			TareaRegistro registro=new TareaRegistro(this);
			registro.start();
			DatagramPacket paqueteRecibido = new DatagramPacket(datosRecibidos, datosRecibidos.length);
		    socket.receive(paqueteRecibido);
		 
		    registro.stop();
		    
		    mensajeSip=SIPMessage.parseMessage(new String(paqueteRecibido.getData()));
		    if (mensajeSip instanceof OKMessage){
		    	System.out.println("Register OK");
		    	enviarACK((OKMessage)mensajeSip);
		    }else if(mensajeSip instanceof NotFoundMessage){
		    	System.out.println("mensaje404 recibido. Error en el registro.");
		    	enviaACK((NotFoundMessage)mensajeSip);
		    	return false;
		    }
			
				
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (SIPException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}
		this.estado=SIN_ESTADO;
		return true;
	}
	private void enviarACK(OKMessage mensajeSip) {
		
		ACKMessage ack=new ACKMessage();
		
		ack.setVias(mensajeSip.getVias() );
		ack.setFromUri(mensajeSip.getFromUri());
		ack.setToName(mensajeSip.getToName());
		ack.setFromName(mensajeSip.getToName());
		ack.setDestination("sip:"+mensajeSip.getToName()+"@"+UA.dominio);
		ack.setToUri(mensajeSip.getToUri());
		ack.setCallId(mensajeSip.getCallId());
		ack.setcSeqNumber("1 REGISTER\n");
		ack.setMaxForwards(MAXFORWARD);
		ack.setContentLength(0);
		
		enviarMensaje(ack);
	}
	private void enviaACK(NotFoundMessage mensajeSip) {
		
		ACKMessage ack=new ACKMessage();
		
		ack.setVias(mensajeSip.getVias() );
		ack.setFromUri(mensajeSip.getFromUri());
		ack.setToName(mensajeSip.getToName());
		ack.setFromName(mensajeSip.getFromName());
		ack.setDestination("sip:"+mensajeSip.getToName()+"@"+UA.dominio);
		ack.setToUri(mensajeSip.getToUri());
		ack.setCallId(mensajeSip.getCallId());
		ack.setcSeqNumber("1 REGISTER\n");
		ack.setMaxForwards(MAXFORWARD);
		ack.setContentLength(0);
		
		enviarMensaje(ack);
		
	}
	public boolean parsearParametros(String[] args){
		if(args.length!=5){
			return false;
		}
		try {
			ipProxy=InetAddress.getByName(args[2]);
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}
		puertoUa=Integer.parseInt(args[1]);
		puertoProxy=Integer.parseInt(args[3]);
		usuarioSip=args[0];
		debug=args[4].equalsIgnoreCase("true");
		return true;
	}
	public void ejecutarOrdenBye() {
		
			if(this.usuarioSIPotro==null){
				System.out.println("Error no hay llamada en curso");
			}else{
				
				double numaleatorio=Math.round(Math.random()*1000000);
	
				ByeMessage mensajeBye;
				ArrayList<String> vias=getViasInicial();
	
				mensajeBye=new ByeMessage();
				
				mensajeBye.setVias(vias);
				mensajeBye.setMaxForwards(MAXFORWARD);
				mensajeBye.setDestination("sip:"+usuarioSIPotro+"@"+dominio);
				mensajeBye.setFromUri("sip:"+usuarioSip);
				mensajeBye.setFromName(usuarioSip.substring(0, usuarioSip.indexOf('@')));
				mensajeBye.setToUri("sip:"+usuarioSIPotro+"@"+dominio);
				mensajeBye.setToName(usuarioSIPotro);
				mensajeBye.setCallId(Double.toString(numaleatorio));
				mensajeBye.setcSeqNumber("3 BYE\n");
				
				mensajeBye.setContentLength(0);
				enviarMensaje(mensajeBye);
				setUsuarSIPotro(null);
			}
	}
	public void setUsuarSIPotro(String usuarioSIPotro) {
		this.usuarioSIPotro=usuarioSIPotro;
		
	}
	
	public void reiniciaScanner() {
		// TODO Auto-generated method stub
		lector.reset();
		//lector=new Scanner(System.in);
	}
	public HiloEscucha getHiloEscucha(){
		return hilosocket;
	}
}
